% *************************************************************************
%                 TWO DIMENSIONAL ELEMENT FREE GALERKIN CODE
%                            Nguyen Vinh Phu
%                        LTDS, ENISE, Juillet 2006
% *************************************************************************

% Description:
% This is a simple Matlab code of the EFG method which is the most familiar
% meshless methods using MLS approximation and Galerkin procedure to derive
% the discrete equations.
% Domain of influence can be : (1) circle and (2) rectangle
% Weight function : cubic or quartic spline function
% Nodes can be uniformly distributed nodes or randomly distributed
% (in the latter case, this is read from a finite element mesh file)
% Numerical integration is done with background mesh
% Essential boundary condition is imposed using Lagrange multipliers or
% penalty method.
% In case of Lagrange multipliers, but use the Dirac delta function to
% approximate lamda, we obtain the boundary point collocation method
% For fracture mechanics, the enriched EFG is used. This is a local PUM
% enrichment scheme which has the form:
%
%           u = N_i * u_i + N_j * H(x) * a_j + N_k * (B_i * b_ik)
%  with H(x) / Heaviside function and B_i(x) the branch functions
%  u_i, a_j and B_ik are nodal parameters.

% External utilities: the mesh, post processing is done with the Matlab
% code of Northwestern university, Illinois, USA

% *************************************************************************

% Problem: the infinite plate with centered circular inclusion
% The material interface is modeled by abs of the signed distance

% +++++++++++++++++++++++++++++++++++++
%          PROBLEM INPUT DATA
% +++++++++++++++++++++++++++++++++++++

clear all
clc
state = 0;
tic;     % help us to see the time required for each step

% Dimension of the domain (it is simply a rectangular region L x W)
L = 10 ;
D = 10 ;

% Material properties

% Matrix
E1  = 1e3 ;
nu1 = 0.3 ;

% Inclusion
E2  = 1e1 ;
nu2 = 0.3 ;

stressState='PLANE_STRAIN';

% Loading
sigmato = 1  ;

% Circular inclusion
r  = 1.0 ;
xc = L/2 ;
yc = D/2 ;

% Inputs special to meshless methods, domain of influence

shape = 'circle' ;         % shape of domain of influence
dmax  = 1.7 ;              % radius = dmax * nodal spacing
form  = 'cubic_spline' ;   % using cubic spline weight function

% Choose method to impose essential boundary condition
% disp_bc_method = 1 : Lagrange multiplier method
% disp_bc_method = 2 : Penalty method

disp_bc_method = 1 ;

% If Lagrange multiplier is used, choose approximation to be used for lamda
% la_approx = 100 ;       % using finite element approximation
% la_approx = 200 ;       % using point collocation method
if disp_bc_method == 1
    la_approx = 200 ;
end

% If penalty function is used, then choose "smart" penalty number :-)
if disp_bc_method == 2
    alpha = 1e7 ;           % penalty number
end

% +++++++++++++++++++++++++++++++++++++
%            NODE GENERATION
% +++++++++++++++++++++++++++++++++++++
% Steps:
%  1. Node coordinates
%  2. Nodes' weight function including : shape (circle or rectangle), size
%  and form (quartic spline or the other)

disp([num2str(toc),'   NODE GENERATION'])

% Node density defined by number of nodes along two directions
nnx = 20 ;
nny = 20  ;

% node generation, node is of size (nnode x 2)
% Corner points of the rectangle domain
% Four corner points
pt1 = [0 0] ;
pt2 = [L 0] ;
pt3 = [L D] ;
pt4 = [0 D] ;
node = square_node_array(pt1,pt2,pt3,pt4,nnx,nny);
numnode = size(node,1);

% Define boundaries
% Bottom, right and top edges are imposed by exact displacement

uln = nnx*(nny-1)+1;       % upper left node number
urn = nnx*nny;             % upper right node number
lrn = nnx;                 % lower right node number
lln = 1;                   % lower left node number
cln = nnx*(nny-1)/2+1;     % node number at (0,0)


topEdge  = [ uln:1:(urn-1); (uln+1):1:urn ]';
botEdge  = [ lln:1:(lrn-1); (lln+1):1:lrn ]';

% GET NODES ON DIRICHLET BOUNDARY AND ESSENTIAL BOUNDARY
botNodes  = unique(botEdge);
topNodes  = unique(topEdge);

dispNodes = [botNodes];
tracNodes = topNodes;

% The bottom edge is constrained along Y direction
% The first node of this edge is constrained to X direction
num_disp_nodes = length(dispNodes)+1;
num_trac_nodes = length(tracNodes)  ;

% Domain of influence for every nodes
% Uniformly distributed nodes
% Definition : rad = dmax*max(deltaX,deltaY)
% deltaX,deltaY are nodal spacing along x,y directions

deltaX = L/(nnx-1);
deltaY = D/(nny-1);
delta  = max(deltaX,deltaY);
di     = ones(1,numnode)*dmax*delta ;

% +++++++++++++++++++++++++++++++++++++
%            GAUSS POINTS
% +++++++++++++++++++++++++++++++++++++
% Steps:
% 1. Mesh generation using available meshing algorithm of FEM
% 2. Using isoparametric mapping to transform Gauss points defined in each
%   element (background element) to global coordinates

disp([num2str(toc),'   BUILD GAUSS POINT FOR DOMAIN '])

% Meshing, Q4 elements
inc_u = 1;
inc_v = nnx;
node_pattern = [ 1 2 nnx+2 nnx+1 ];
element = make_elem(node_pattern,nnx-1,nny-1,inc_u,inc_v);

% Building Gauss points
[w,q]=quadrature(6, 'GAUSS', 2 );  % 4x4 Gaussian quadrature for each cell
W = [];
Q = [];
J = [];
for e = 1 : size(element,1)        % element loop
    sctr=element(e,:);             % element scatter vector
    for i = 1:size(w,1)                        % quadrature loop
        pt=q(i,:);                             % quadrature point
        wt=w(i);                               % quadrature weight
        [N,dNdxi]=lagrange_basis('Q4',pt);     % Q4 shape functions
        J0=node(sctr,:)'*dNdxi;                % element Jacobian matrix
        Q = [Q; N' * node(sctr,:)];            % global Gauss point
        W = [W; wt];                           % global Gauss point
        J = [J; det(J0)];
    end
end
clear w,q ;

% +++++++++++++++++++++++++++++++++++++
%        LEVEL SET INITIALIZATION
%      SELECTION OF ENRICHED NODES
% +++++++++++++++++++++++++++++++++++++


for i = 1 : numnode
    x = node(i,1);
    y = node(i,2);
    d = sqrt((x-xc)^2+(y-yc)^2);
    ls(i)  = d - (r + di(i));
end

enrich_node = find(ls < 0);

% +++++++++++++++++++++++++++++++++++++
%  PLOT NODES,BACKGROUND MESH, GPOINTS
% +++++++++++++++++++++++++++++++++++++
disp([num2str(toc),'   PLOT NODES AND GAUSS POINTS'])
figure
hold on
cntr = plot([0,L,L,0,0],[0,0,D,D,0]);
set(cntr,'LineWidth',3)
h = plot(node(:,1),node(:,2),'bo');
set(h,'MarkerSize',8.5);

theta = -pi:0.1:pi;
x = xc + r*cos(theta);
y = yc + r*sin(theta);
cir = plot(x,y,'k-');
set(cir,'LineWidth',2)

n1 = plot(node(enrich_node,1),node(enrich_node,2),'r*');
set(n1,'MarkerSize',15);
plot(node(dispNodes,1),node(dispNodes,2),'ks');
axis off
axis equal


% +++++++++++++++++++++++++++++++++++++
%          DOMAIN ASSEMBLY
% +++++++++++++++++++++++++++++++++++++
disp([num2str(toc),'   DOMAIN ASSEMBLY'])

% Initialisation

total_num_node = numnode + size(enrich_node,2)*1  ;
K = sparse(2*total_num_node,2*total_num_node);
f = zeros(2*total_num_node,1);

% Due to the presence of additional dofs, the assembly is a little
% bit difficult than in FEM. We use fictitious nodes to handle these
% additional dofs. At an enriched node, we add one fantom node. These fictitious nodes
% are numbered from the total number of true nodes, ie, from numnode+1 ...

pos = zeros(numnode,1);
nsnode = 0 ;
for i = 1 : numnode
    [snode] = ismember(i,enrich_node);
    if (snode ~= 0)
        pos(i) = (numnode + nsnode*1 ) + 1 ;
        nsnode = nsnode + 1 ;
    end
end


% --------------------------
%    Loop on Gauss points
% --------------------------

for igp = 1 : size(W,1)
    pt = Q(igp,:);                             % quadrature point

    % ----------------------------------------------
    % find nodes in neighbouring of Gauss point pt
    % ----------------------------------------------
    [index] = define_support(node,pt,di);

    % For enriched EFG, here is the most important part
    % Nodes within neighbouring of GPs include normal nodes, H enriched
    % nodes and tip enriched nodes
    % ISMEMBER(A,S) for the array A returns an array of the same size as A
    % containing 1 where the elements of A are in the set S and 0 otherwise.

    [snode,s_loc] = ismember(index,enrich_node);
   
    % number of H and tip enriched nodes
    % in index
    num_enr_node = size(find(snode ~= 0),2);
  
    % Scatter of standard part of B matrix
    le   = length(index);
    sctrStdB = zeros(2*le,1);
    sctrStdB(1:2:2*le) = index.*2-1 ; % x displacement
    sctrStdB(2:2:2*le) = index.*2   ; % y displacement

    % No enriched nodes in index
    if (num_enr_node == 0) 
        sctrB = sctrStdB ;
    % There is at least one enriched node in index
    else
        sn = num_enr_node;
        sctrEnrB = zeros(2*(sn*1),1);
        cnt = 0 ;
        for k = 1 : le
            nodeI = index(k) ; % I-th node in index
            if (snode(k) ~= 0)
                cnt = cnt + 1 ;
                sctrEnrB(2*cnt - 1) = 2 * pos(nodeI) - 1;
                sctrEnrB(2*cnt    ) = 2 * pos(nodeI)    ;
            end
        end
        sctrB = [ sctrStdB;sctrEnrB ];
    end
   
    %  compute B matrix
    [phi,dphidx,dphidy] = MLS_ShapeFunction(pt,index,node,di,form);

    StdB = zeros(3,2*le);
    StdB(1,1:2:2*le)  = dphidx ;
    StdB(2,2:2:2*le)  = dphidy ;
    StdB(3,1:2:2*le)  = dphidy ;
    StdB(3,2:2:2*le)  = dphidx ;

    % no enriched nodes, B is simply standard B matrix
    if (num_enr_node == 0)
        B = StdB;
    % there are enriched nodes, compute enriched part, Benr
    else
        Benr = [];
        for m = 1 : le
            if (snode(m) ~= 0)
                % compute the enrichment function and derivatives
                x = pt(1);
                y = pt(2);                
                d = sqrt((x-xc)^2+(y-yc)^2);
                F    = abs(d - r);
                dFdx = sign(F)*(x-xc)/d;
                dFdy = sign(F)*(y-yc)/d;                
                
                % B matrix of node m
                aa = dphidx(m)*F + phi(m)*dFdx ;
                bb = dphidy(m)*F + phi(m)*dFdy ;
                BI_enr = [aa 0 ; 0 bb ; bb aa];
                % Add to the total enriched B matrix
                Benr   = [Benr BI_enr];
                clear BI_enr ;
            end
        end   
        
        % Total B matrix
        B = [StdB Benr];
        clear StdB; clear Benr;
    end     
    
    % Compute the compliance matrix C
    x = pt(1);
    y = pt(2);
    d = sqrt((x-xc)^2+(y-yc)^2);
    
    if d < r 
        E  = E2 ;
        nu = nu2; 
    else
        E  = E1 ;
        nu = nu1; 
    end
    
    if ( strcmp(stressState,'PLANE_STRESS') )
        C=E/(1-nu^2)*[ 1   nu 0;
            nu  1  0 ;
            0   0  0.5*(1-nu) ];
    else
        C=E/(1+nu)/(1-2*nu)*[ 1-nu  nu  0;
            nu    1-nu 0;
            0     0  0.5-nu ];
    end
    
    % Then K matrix
    K(sctrB,sctrB) = K(sctrB,sctrB) + B'*C*B * W(igp)*J(igp) ;
end          % end of loop on Gauss Points


% +++++++++++++++++++++++++++++++++++++
%    INTEGRATE ON TRACTION BOUNDARY
% +++++++++++++++++++++++++++++++++++++
% Steps:
% 1. Build up Gauss points on the boundary, 4 GPs are used
% 2. Loop on these GPs to form the f vector
disp([num2str(toc),'   INTEGRATION ON TRACTION BOUNDARY'])

% --------------------------------------
%   Generation of Gauss points
% --------------------------------------

% 4 point quadrature for each "1D element"
[W1,Q1] = quadrature(4, 'GAUSS', 1 );

Wt = [];
Qt = [];
Jt = [];

for i = 1 : (num_trac_nodes - 1)
    sctr = [tracNodes(i) tracNodes(i+1)];
    for q = 1:size(W1,1)
        pt = Q1(q,:);
        wt = W1(q);
        [N,dNdxi] = lagrange_basis('L2',pt);
        J0=dNdxi'*node(sctr,:);
        Qt = [Qt; N' * node(sctr,:)];
        Wt = [Wt; wt];
        Jt = [Jt; norm(J0)];
    end % of quadrature loop
end

% ----------------------------------
%   Form the nodal force vector f
% ----------------------------------

for igp = 1 : size(Wt,1)
    pt = Qt(igp,:);                             % quadrature point
    wt = Wt(igp);                               % quadrature weight
    detJ   = Jt(igp);
    [index] = define_support(node,pt,di);
    sctry = index.*2 ;
    [phi,dphidx,dphidy] = MLS_ShapeFunction(pt,index,node,di,form);
    f(sctry) = f(sctry) + phi'*sigmato*detJ*wt;
end


% +++++++++++++++++++++++++++++++++++++
%   INTEGRATE ON DISPLACEMENT BOUNDARY
% +++++++++++++++++++++++++++++++++++++
% Steps:
% 1. Build up Gauss points on the boundary, 4 GPs are used
% 2. Loop on these GPs to form the vector q and matrix G

disp([num2str(toc),'   INTEGRATION ON DISPLACEMENT BOUNDARY'])


% If Lagrange multiplier method is used
if disp_bc_method == 1
    disp(' Lagrange multiplier method is used ')
    % --------------------------------------
    %       Form vector q and matrix G
    % --------------------------------------
    qk = zeros(1,2*num_disp_nodes);
    G  = zeros(2*total_num_node,2*num_disp_nodes);

    % If point collocation method is used. Assume that these collocation
    % points are coincident to nodes on displacement boundary
    if (la_approx == 200)
        disp(' Point collocation method is used ')
        % for the bottom edge
        for i = 1 : num_disp_nodes-1      % loop on collocation points
            pt = node(dispNodes(i),:);
            % S matrix
            S = [0 0; 0 1]; % Y direction
            % G matrix
            [index] = define_support(node,pt,di);
            [phi,dphidx,dphidy] = MLS_ShapeFunction(pt,index,node,di,form);
            for j = 1 : size(index,2)
                row1 = 2*index(j)-1  ;
                row2 = 2*index(j)    ;
                Gj   =  -phi(j) * S ;
                G(row1:row2,2*i-1:2*i) = G(row1:row2,2*i-1:2*i) + Gj ;
            end
        end
        % for the left lower corner node, x direction
        pt = node(1,:);
        i  = i+1;
        % S matrix
        S = [1 0; 0 0]; % X direction
        % G matrix
        [index] = define_support(node,pt,di);
        [phi,dphidx,dphidy] = MLS_ShapeFunction(pt,index,node,di,form);
        for j = 1 : size(index,2)
            row1 = 2*index(j)-1  ;
            row2 = 2*index(j)    ;
            Gj   =  -phi(j) * S ;
            G(row1:row2,2*i-1:2*i) = G(row1:row2,2*i-1:2*i) + Gj ;
        end
    end   % end of if boundary point collocation is used
end       % end of if disp_bc_method = 'Lagrange'


% +++++++++++++++++++++++++++++++++++++
%      SOLUTION OF THE EQUATIONS
% +++++++++++++++++++++++++++++++++++++

disp([num2str(toc),'   SOLUTION OF EQUATIONS'])


% If Lagrange multiplier is used then, extend the unknowns u by lamda
if disp_bc_method == 1
    f = [f;qk'];                                % f = {f;qk}
    m = ([K G; G' zeros(num_disp_nodes*2)]);    % m = [K GG;GG' 0]
    d = m\f;                                    % d = {u;lamda}
    % just get nodal parameters u_i, not need Lagrange multipliers
    u = d(1:2*total_num_node);
end

clear d ;

% +++++++++++++++++++++++++++++++++++++
%    COMPUTE THE TRUE DISPLACEMENTS
% +++++++++++++++++++++++++++++++++++++

disp([num2str(toc),'   INTERPOLATION TO GET TRUE DISPLACEMENT'])

for i = 1 : numnode
    pt = node(i,:);
    [index] = define_support(node,pt,di);
    le      = length(index);

    [snode,s_loc] = ismember(index,enrich_node);
    if (any(snode) == 0)
        no_enr_node  = 1 ; 
        num_enr_node = 0 ;
    else
        no_enr_node = 0;
        num_enr_node = size(find(snode ~= 0),2);
    end

    % shape function at nodes in neighbouring of node i
    [phi,dphidx,dphidy] = MLS_ShapeFunction(pt,index,node,di,form);

    % Recall the enriched displacement approximation:
    % u(x) = Ni*ui + Nj*H*bj 

    % The first part Ni*ui
    stdUx = phi*u(index.*2-1);
    stdUy = phi*u(index.*2);

    if (no_enr_node == 1) 
        ux(i) = stdUx ;
        uy(i) = stdUy ;
    else
        % The second part exists only with enriched particles
        FUx = 0 ;
        FUy = 0 ;
        for m = 1 : le
            % a split enriched node
            if (snode(m) ~= 0)
                % compute F function
                x = pt(1);
                y = pt(2);
                d = sqrt((x-xc)^2+(y-yc)^2);
                F    = abs(d - r);
                FUx = FUx + phi(m)*F*u(2 * pos(index(m)) - 1) ;
                FUy = FUy + phi(m)*F*u(2 * pos(index(m))    ) ;
            end
        end   % end of loop on nodes in neighbour of Gp
        ux(i) = stdUx + FUx  ;
        uy(i) = stdUy + FUy ;
    end
end

% ++++++++++++++++++++++++++++++++++++
%    COMPUTE STRESSES AT PARTICLES
% ++++++++++++++++++++++++++++++++++++
disp([num2str(toc),'   STRESSES COMPUTATION'])

ind = 0 ;
for igp = 1 : size(node,1)
    pt = node(igp,:);
    [index] = define_support(node,pt,di);

    [snode,s_loc] = ismember(index,enrich_node);
   
    % number of enriched nodes in index
    num_enr_node = size(find(snode ~= 0),2);

    % Scatter of standard part of B matrix
    le   = length(index);
    sctrStdB = zeros(2*le,1);
    sctrStdB(1:2:2*le) = index.*2-1 ; % x displacement
    sctrStdB(2:2:2*le) = index.*2   ; % y displacement

    % No enriched nodes in index
    if (num_enr_node == 0)
        sctrB = sctrStdB ;
        % There is at least one enriched node in index
    else
        sn = num_enr_node;
        sctrEnrB = zeros(2*(sn*1),1);
        cnt = 0 ;
        for k = 1 : le
            nodeI = index(k) ; % I-th node in index
            if (snode(k) ~= 0)
                cnt = cnt + 1 ;
                sctrEnrB(2*cnt - 1) = 2 * pos(nodeI) - 1;
                sctrEnrB(2*cnt    ) = 2 * pos(nodeI)    ;
            end
        end
        sctrB = [ sctrStdB;sctrEnrB ];
    end

    %  compute B matrix
    [phi,dphidx,dphidy] = MLS_ShapeFunction(pt,index,node,di,form);

    StdB = zeros(3,2*le);
    StdB(1,1:2:2*le)  = dphidx ;
    StdB(2,2:2:2*le)  = dphidy ;
    StdB(3,1:2:2*le)  = dphidy ;
    StdB(3,2:2:2*le)  = dphidx ;

    % no enriched nodes, B is simply standard B matrix
    if (num_enr_node == 0)
        B = StdB;
        % there are enriched nodes, compute enriched part, Benr
    else
        Benr = [];
        for m = 1 : le
            if (snode(m) ~= 0)
                % compute the enrichment function and derivatives
                x = pt(1);
                y = pt(2);
                d = sqrt((x-xc)^2+(y-yc)^2);
                F    = abs(d - r);
                dFdx = sign(F)*(x-xc)/d;
                dFdy = sign(F)*(y-yc)/d;

                % B matrix of node m
                aa = dphidx(m)*F + phi(m)*dFdx ;
                bb = dphidy(m)*F + phi(m)*dFdy ;
                BI_enr = [aa 0 ; 0 bb ; bb aa];
                % Add to the total enriched B matrix
                Benr   = [Benr BI_enr];
                clear BI_enr ;
            end
        end

        % Total B matrix
        B = [StdB Benr];
        clear StdB; clear Benr;
    end

    % Compute the compliance matrix C
    x = pt(1);
    y = pt(2);
    d = sqrt((x-xc)^2+(y-yc)^2);

    if d < r
        E  = E2 ;
        nu = nu2;
    else
        E  = E1 ;
        nu = nu1;
    end

    if ( strcmp(stressState,'PLANE_STRESS') )
        C=E/(1-nu^2)*[ 1   nu 0;
            nu  1  0 ;
            0   0  0.5*(1-nu) ];
    else
        C=E/(1+nu)/(1-2*nu)*[ 1-nu  nu  0;
            nu    1-nu 0;
            0     0  0.5-nu ];
    end

    ind = ind + 1 ;
    stress_gp(1:3,ind) = C*B*u(sctrB);  % sigma = C*epsilon = C*(B*u)
end

fac = 0 ; % visualization factor
figure
plot_field(node+fac*[ux' uy'],element,'Q4',stress_gp(2,:));
axis('equal');
xlabel('X');
ylabel('Y');
title('\sigma_{xx}');
set(gcf,'color','white');
colorbar('vert');
title('EFG stress, \sigma_{xx}')

% ++++++++++++++++++
%    VISUALIATION
% ++++++++++++++++++

disp([num2str(toc),'   DEFORMED CONFIGURATION'])

% ----------------------------------
fac = 550 ; % visualization factor
figure
hold on
h = plot(node(:,1)+fac*ux',...
    node(:,2)+fac*uy','r*');
set(h,'MarkerSize',7);
title('Numerical deformed shape')
axis equal
% ----------------------------------



