% *************************************************************************
%                 TWO DIMENSIONAL ELEMENT FREE GALERKIN CODE
%                            Nguyen Vinh Phu
%                        LTDS, ENISE, Juillet 2006
% *************************************************************************

% Description:
% This is a simple Matlab code of the EFG method which is the most familiar 
% meshless methods using MLS approximation and Galerkin procedure to derive
% the discrete equations.
% Domain of influence can be : (1) circle and (2) rectangle
% Weight function : cubic or quartic spline function
% Nodes can be uniformly distributed nodes or randomly distributed 
% (in the latter case, this is read from a finite element mesh file)
% Numerical integration is done with background mesh
% Essential boundary condition is imposed using Lagrange multipliers or
% penalty method. 
% In case of Lagrange multipliers, but use the Dirac delta function to
% approximate lamda, we obtain the boundary point collocation method

% External utilities: the mesh, post processing is done with the Matlab
% code of Northwestern university, Illinois, USA

% *************************************************************************

% Problem: the Timoshenko beam problem as given in EFG paper

% +++++++++++++++++++++++++++++++++++++
%          PROBLEM INPUT DATA
% +++++++++++++++++++++++++++++++++++++

clear all
clc
state = 0;
tic;           % help us to see the time required for each step

% Dimension of the domain (it is simply a rectangular region L x W)
L = 48 ;
D = 12 ;

% Material properties
E  = 30e6 ;
nu = 0.3 ;

% Chargement
P = 1000 ;

% Inputs special to meshless methods, domain of influence

shape = 'circle' ;         % shape of domain of influence
dmax  = 1.5 ;              % radius = dmax * nodal spacing
form  = 'cubic_spline' ;   % using cubic spline weight function

% Choose method to impose essential boundary condition
% disp_bc_method = 1 : Lagrange multiplier method
% disp_bc_method = 2 : Penalty method

disp_bc_method = 1 ;

% If Lagrange multiplier is used, choose approximation to be used for lamda
% la_approx = 100 ;       % using finite element approximation
% la_approx = 200 ;       % using point collocation method    
if disp_bc_method == 1
    la_approx = 100 ;        
end

% If penalty function is used, then choose "smart" penalty number :-)
if disp_bc_method == 2
    alpha = 1e4 ;           % penalty number
end

% +++++++++++++++++++++++++++++++++++++
%            NODE GENERATION
% +++++++++++++++++++++++++++++++++++++
% Steps:
%  1. Node coordinates
%  2. Nodes' weight function including : shape (circle or rectangle), size
%  and form (quartic spline or the other)

disp([num2str(toc),'   NODE GENERATION'])

% Node density defined by number of nodes along two directions
nnx = 11 ;
nny = 7  ;

% node generation, node is of size (nnode x 2)
node = square_node_array([0 -D/2],[L -D/2],[L D/2],[0 D/2],nnx,nny);
numnode = size(node,1);

% Boundary nodes
% 1. Displacement condition on the left edge with x = 0
% 2. Traction condition on the right edge with x = L

disp_nodes = find(node(:,1)==0) ;
trac_nodes = find(node(:,1)==L) ;
num_disp_nodes = length(disp_nodes);
num_trac_nodes = length(trac_nodes);

ubar = zeros(2*num_disp_nodes,1); % initialize of vector u bar
f    = zeros(2*numnode,1);        % nodal force vector

% Domain of influence for every nodes 
% Uniformly distributed nodes
% Definition : rad = dmax*max(deltaX,deltaY)
% deltaX,deltaY are nodal spacing along x,y directions

deltaX = L/(nnx-1);
deltaY = D/(nny-1);
delta  = max(deltaX,deltaY);
di     = ones(1,numnode)*dmax*delta ;

% +++++++++++++++++++++++++++++++++++++
%            GAUSS POINTS
% +++++++++++++++++++++++++++++++++++++
% Steps:
% 1. Mesh generation using available meshing algorithm of FEM
% 2. Using isoparametric mapping to transform Gauss points defined in each
%   element (background element) to global coordinates

disp([num2str(toc),'   BUILD GAUSS POINT FOR DOMAIN '])

% Meshing, Q4 elements
inc_u = 1;
inc_v = nnx;
node_pattern = [ 1 2 nnx+2 nnx+1 ];
element = make_elem(node_pattern,nnx-1,nny-1,inc_u,inc_v);

% Building Gauss points
[w,q]=quadrature(4, 'GAUSS', 2 );  % 4x4 Gaussian quadrature for each cell
W = [];
Q = [];
J = [];
for e = 1 : size(element,1)        % element loop
    sctr=element(e,:);             % element scatter vector
    for i = 1:size(w,1)                        % quadrature loop
        pt=q(i,:);                             % quadrature point
        wt=w(i);                               % quadrature weight
        [N,dNdxi]=lagrange_basis('Q4',pt);     % Q4 shape functions
        J0=node(sctr,:)'*dNdxi;                % element Jacobian matrix
        Q = [Q; N' * node(sctr,:)];            % global Gauss point
        W = [W; wt];                           % global Gauss point
        J = [J; det(J0)];
    end
end
clear w,q ;

% +++++++++++++++++++++++++++++++++++++
%  PLOT NODES,BACKGROUND MESH, GPOINTS
% +++++++++++++++++++++++++++++++++++++
disp([num2str(toc),'   PLOT NODES AND GAUSS POINTS'])
figure
hold on
cntr = plot([0,L,L,0,0],[-D/2,-D/2,D/2,D/2,-D/2]);
set(cntr,'LineWidth',3)
h = plot(node(:,1),node(:,2),'bo');
set(h,'MarkerSize',10.5);
axis([-1 L -D D])
axis off

figure
hold on
cntr = plot([0,L,L,0,0],[-D/2,-D/2,D/2,D/2,-D/2]);
set(cntr,'LineWidth',3)
plot_mesh(node,element,'Q4','--');
%plot(Q(:,1),Q(:,2),'r*');
axis off

% +++++++++++++++++++++++++++++++++++++
%          DOMAIN ASSEMBLY
% +++++++++++++++++++++++++++++++++++++
disp([num2str(toc),'   DOMAIN ASSEMBLY'])

% Initialisation
K = sparse(2*numnode,2*numnode);
f = zeros(2*numnode,1);

% Plane stress matrix
C = (E/(1-nu*nu)) * [1 nu 0 ;
                     nu 1 0
                     0 0 0.5*(1-nu)] ;
% ----------------------------------------------
%             Loop on Gauss points
% ----------------------------------------------
for igp = 1 : size(W,1)
    pt = Q(igp,:);                             % quadrature point
    wt = W(igp);                               % quadrature weight

    % ----------------------------------------------
    % find nodes in neighbouring of Gauss point pt
    % ----------------------------------------------
    [index] = define_support(node,pt,di);

    % --------------------------------------
    %   compute B matrix, K matrix
    % --------------------------------------
    % Loop on nodes within the support, compute dPhi of each node
    % Then get the B matrix
    % Finally assemble to the K matrix

    B = zeros(3,2*size(index,2)) ;
    en = zeros(1,2*size(index,2));  
    [phi,dphidx,dphidy] = MLS_ShapeFunction(pt,index,node,di,form);
    for m = 1 : size(index,2)      
        B(1:3,2*m-1:2*m) = [dphidx(m) 0 ; 
                           0 dphidy(m);
                           dphidy(m) dphidx(m)];
        en(2*m-1) = 2*index(m)-1;
        en(2*m  ) = 2*index(m)  ;
    end
    K(en,en) = K(en,en) + B'*C*B * W(igp)*J(igp) ;
end

% +++++++++++++++++++++++++++++++++++++
%    INTEGRATE ON TRACTION BOUNDARY
% +++++++++++++++++++++++++++++++++++++
% Steps:
% 1. Build up Gauss points on the boundary, 4 GPs are used
% 2. Loop on these GPs to form the f vector
disp([num2str(toc),'   INTEGRATION ON TRACTION BOUNDARY'])

% --------------------------------------
%   Generation of Gauss points
% --------------------------------------

% 4 point quadrature for each "1D element"
[W1,Q1]=quadrature(4, 'GAUSS', 1 ); 

Wt = [];
Qt = [];
Jt = [];

for i = 1 : (num_trac_nodes - 1)
    sctr = [trac_nodes(i) trac_nodes(i+1)]; 
    for q = 1:size(W1,1)                            
        pt = Q1(q,:);                               
        wt = W1(q);                                 
        [N,dNdxi] = lagrange_basis('L2',pt);        
        J0=dNdxi'*node(sctr,:);                    
        Qt = [Qt; N' * node(sctr,:)];              
        Wt = [Wt; wt];                             
        Jt = [Jt; norm(J0)];
    end % of quadrature loop
end

% --------------------------------------
%   Form the nodal force vector f
% --------------------------------------
I = (1/12)*D^3;

for igp = 1 : size(Wt,1)
    pt = Qt(igp,:);                             % quadrature point
    wt = Wt(igp);                               % quadrature weight
    [index] = define_support(node,pt,di);
    en = zeros(1,2*size(index,2));              
    force = zeros(1,2*size(index,2));
    [phi,dphidx,dphidy] = MLS_ShapeFunction(pt,index,node,di,form);
    tx = 0.0 ;
    ty = -(P/(2*I))*((D*D)/4-pt(1,2)^2);
    for j = 1 : size(index,2)
        en(2*j-1) = 2*index(j)-1;
        en(2*j  ) = 2*index(j)  ;
        force(2*j-1) = tx*phi(j);
        force(2*j  ) = ty*phi(j);
    end
    f(en) = f(en) + Jt(igp) * wt * force' ;
end

% +++++++++++++++++++++++++++++++++++++
%   INTEGRATE ON DISPLACEMENT BOUNDARY
% +++++++++++++++++++++++++++++++++++++
% Steps:
% 1. Build up Gauss points on the boundary, 4 GPs are used
% 2. Loop on these GPs to form the vector q and matrix G 

disp([num2str(toc),'   INTEGRATION ON DISPLACEMENT BOUNDARY'])

% --------------------------------------
%   Generation of Gauss points
% --------------------------------------
Wu = [];
Qu = [];
Ju = [];

for i = 1 : (num_disp_nodes - 1)
    sctr = [disp_nodes(i) disp_nodes(i+1)]; 
    for q = 1:size(W1,1)                            
        pt = Q1(q,:);                               
        wt = W1(q);                                 
        [N,dNdxi] = lagrange_basis('L2',pt);       
        J0=dNdxi'*node(sctr,:);                     
        Qu = [Qu; N' * node(sctr,:)];               
        Wu = [Wu; wt];                              
        Ju = [Ju; norm(J0)];                       
    end % of quadrature loop
end

% If Lagrange multiplier method is used
if disp_bc_method == 1
    disp(' Lagrange multiplier method is used ')
    % --------------------------------------
    %       Form vector q and matrix G
    % --------------------------------------
    qk = zeros(1,2*num_disp_nodes);
    G  = zeros(2*numnode,2*num_disp_nodes);
        
    % If point collocation method is used. Assume that these collocation
    % points are coincident to nodes on displacement boundary
    if (la_approx == 200)  
        disp(' Point collocation method is used ')
        for i = 1 : num_disp_nodes      % loop on collocation points
            pt = node(disp_nodes(i),:);                           
        
            % compute exact displacement of Timoshenko problem, u_bar
            x = pt(1,1) ; y = pt(1,2) ;
            fac = P/(6*E*I);
            ux_exact =  fac*y*((6*L-3*x)*x+(2+nu)*(y^2-D*D/4));
            uy_exact = -fac*((L-x)*3*nu*y^2+0.25*(4+5*nu)*D*D*x+(3*L-x)*x^2);
            
            % qk vector 
            qk(1,2*i-1) = -ux_exact ;
            qk(1,2*i)   = -uy_exact ;
            
            % S matrix
            S = [1 0; 0 1]; % both directions
            % G matrix
            [index] = define_support(node,pt,di);
            [phi,dphidx,dphidy] = MLS_ShapeFunction(pt,index,node,di,form);
            for j = 1 : size(index,2)
                row1 = 2*index(j)-1 ;
                row2 = 2*index(j)   ;        
                Gj   =  -phi(j) * S ; 
                G(row1:row2,2*i-1:2*i) = G(row1:row2,2*i-1:2*i) + Gj ;
                                        
            end
        end
    end   % end of if boundary point collocation is used
    
    % if Lagrange multipliers are interpolated by finite elements 
    % along the displacement boundary 
    if (la_approx == 100)
        disp(' Finite element interpolation for ML is used ')
        m1 = 0 ; 
        for i = 1 : (num_disp_nodes - 1)
            sctr = [disp_nodes(i) disp_nodes(i+1)];
            m1 = m1 + 1 ;
            m2 = m1 + 1 ;
            %le = norm(disp_nodes(m1) - disp_nodes(m2));
            le = norm(node(sctr(1),:) - node(sctr(2),:));
            for q = 1:size(W1,1)
                pt = Q1(q,:);
                wt = W1(q);
                [N,dNdxi] = lagrange_basis('L2',pt);
                J0 = dNdxi'*node(sctr,:);
                detJ = norm(J0) ;
                pt = N' * node(sctr,:); % global GP
                % compute exact displacement of Timoshenko problem, u_bar
                x = pt(1,1) ; y = pt(1,2) ;
                fac = P/(6*E*I);
                ux_exact =  fac*y*((6*L-3*x)*x+(2+nu)*(y^2-D*D/4));
                uy_exact = -fac*((L-x)*3*nu*y^2+0.25*(4+5*nu)*D*D*x+(3*L-x)*x^2);

                N1 = 1 - pt(1,2)/le ; N2 = 1 - N1 ;
                % qk vector
                qk(2*m1-1) = qk(2*m1-1) - wt * detJ * N1 * ux_exact ;
                qk(2*m1)   = qk(2*m1)   - wt * detJ * N1 * uy_exact ;
                qk(2*m2-1) = qk(2*m2-1) - wt * detJ * N2 * ux_exact ;
                qk(2*m2)   = qk(2*m2)   - wt * detJ * N2 * uy_exact ;

                % G matrix
                [index] = define_support(node,pt,di);
                [phi,dphidx,dphidy] = MLS_ShapeFunction(pt,index,node,di,form);
                for j = 1 : size(index,2)
                    row1 = 2*index(j)-1 ;
                    row2 = 2*index(j)   ;
                    G1   = - wt * detJ * phi(j) * [N1 0 ; 0 N1];
                    G2   = - wt * detJ * phi(j) * [N2 0 ; 0 N2];
                    G(row1:row2,2*m1-1:2*m1) = G(row1:row2,2*m1-1:2*m1) + G1;
                    G(row1:row2,2*m2-1:2*m2) = G(row1:row2,2*m2-1:2*m2) + G2;
                end
            end
        end
    end
end       % end of if disp_bc_method = 'Lagrange'

% If penalty method is use, then one must modify the stiffness matrix and
% the nodal force vector
% $K_{ij}$ = Kij - alpha \int phi_i phi_j d \gamma_u
% fj  = fj - alpha \int phi_i u_bar d \gamma_u

if disp_bc_method == 2
    fu = zeros(2*numnode,1);
    k  = zeros(2*numnode,2*numnode);
    for igp = 1 : size(Wu,1)
        pt = Qu(igp,:);                             % quadrature point
        wt = Wu(igp);                               % quadrature weight
        [index] = define_support(node,pt,di);
        [phi,dphidx,dphidy] = MLS_ShapeFunction(pt,index,node,di,form);
        % compute exact displacement of Timoshenko problem
        x = pt(1,1) ; y = pt(1,2) ;
        fac = P/(6*E*I);
        ux_exact =  fac*y*((6*L-3*x)*x+(2+nu)*(y^2-D*D/4));
        uy_exact = -fac*((L-x)*3*nu*y^2+0.25*(4+5*nu)*D*D*x+(3*L-x)*x^2); 
        le = size(index,2);
        en = zeros(1,2*le);
        force = zeros(1,2*le);
        S = [1 0 ; 0 1];
        for j = 1 : le
            en(2*j-1) = 2*index(j)-1;
            en(2*j  ) = 2*index(j)  ;
            force(2*j-1) = phi(j)*ux_exact;
            force(2*j  ) = phi(j)*uy_exact;
        end
        fu(en) = fu(en)     + Jt(igp) * wt * force' ;
        k(en,en) = k(en,en) + Jt(igp) * wt * phi*phi'*S ; 
    end % end of loop on Gauss points
    f = f - alpha*fu;    
    K = K - alpha*k;
end

% +++++++++++++++++++++++++++++++++++++
%      SOLUTION OF THE EQUATIONS
% +++++++++++++++++++++++++++++++++++++

disp([num2str(toc),'   SOLUTION OF EQUATIONS'])

% If penalty function is used, then the unknowns u is the same, just solve
% to get it:-)
if disp_bc_method == 2
    d = K\f ;
    u2 = d;
    for i = 1 : numnode
        u(1,i) = d(2*i-1); % x displacement
        u(2,i) = d(2*i);   % y displacement
    end
end

% If Lagrange multiplier is used then, extend the unknowns u by lamda
if disp_bc_method == 1
    f = [f;qk'];                                % f = {f;qk}
    m = ([K G; G' zeros(num_disp_nodes*2)]);    % m = [K GG;GG' 0] 
    d = m\f;                                    % d = {u;lamda}
    u2 = d(1:numnode*2);
    % just get displacement parameters u_i
    for i = 1 : numnode
        u(1,i) = d(2*i-1); % x displacement
        u(2,i) = d(2*i);   % y displacement
    end
end

clear d ;

% +++++++++++++++++++++++++++++++++++++
%    COMPUTE THE TRUE DISPLACEMENTS
% +++++++++++++++++++++++++++++++++++++
disp([num2str(toc),'   INTERPOLATION TO GET TRUE DISPLACEMENT'])
disp = zeros(1,2*numnode);
for i = 1 : numnode
    [index] = define_support(node,node(i,:),di);
    % shape function at nodes in neighbouring of node i
    [phi,dphidx,dphidy] = MLS_ShapeFunction(node(i,:),index,node,di,form);
    disp(1,2*i-1) = phi*u(1,index)'; % x nodal displacement
    disp(1,2*i)   = phi*u(2,index)'; % y nodal displacement
end
clear u;
% +++++++++++++++++++++++++++++++++++++
%   COMPUTE STRESSES AT GAUSS POINTS
% +++++++++++++++++++++++++++++++++++++
disp([num2str(toc),'   COMPUTE STRESS AT GAUSS POINTS'])
ind = 0 ;
for igp = 1 : size(W,1)
    pt = Q(igp,:);                             % quadrature point
    wt = W(igp);                               % quadrature weight
    [index] = define_support(node,pt,di);
    B = zeros(3,2*size(index,2)) ;
    en = zeros(1,2*size(index,2));  
    [phi,dphidx,dphidy] = MLS_ShapeFunction(pt,index,node,di,form);
    for m = 1 : size(index,2)       
        B(1:3,2*m-1:2*m) = [dphidx(m) 0 ; 
                            0 dphidy(m); 
                            dphidy(m) dphidx(m)];
        en(2*m-1) = 2*index(m)-1;
        en(2*m  ) = 2*index(m)  ;
    end
    ind = ind + 1 ;
    stress_gp(1:3,ind) = C*B*u2(en);  % sigma = C*epsilon = C*(B*u)
end

% +++++++++++++++++++++++++++++++++++++
%            VISUALIATION
% +++++++++++++++++++++++++++++++++++++

% Deformed configuration
% ----------------------

figure
plot_mesh(node,element,'Q4','-');
fac = 200 ; % visualization factor
hold on
h = plot(node(:,1)+fac*disp(1,1:2:2*numnode-1)',...
    node(:,2)+fac*disp(1,2:2:2*numnode)','b*');
set(h,'MarkerSize',7);
axis off

% Stress visualization
% ----------------------
% Stresses are computed at Gauss points
% Gauss points are then used to build a Delaunay triangulation, then plot 
figure
hold on
tri = delaunay(Q(:,1),Q(:,2));
plot_field(Q,tri,'T3',stress_gp(1,:));
axis('equal');
xlabel('X');
ylabel('Y');
title('Sigma XX');
set(gcf,'color','white');
colorbar('vert');
opts = struct('Color','rgb','Bounds','tight');
% exportfig(gcf,'beam_stress_x.eps',opts)

% Export to Tecplot for better plot
varname{1}='X';
varname{2}='Y';
varname{3}='sigma_xx';
varname{4}='sigma_yy';
varname{5}='sigma_xy';
%varname{6}='sigma_vonMises';
sigma=[stress_gp(1,:)' stress_gp(2,:)' stress_gp(3,:)'];
value=[Q(:,1) Q(:,2) sigma];
% tecplotout(tri,'T3',Q,varname,value,'beam_stress.dat');

for igp = 1 : size(W,1)
    pt = Q(igp,:);                             % quadrature point
    x = pt(1,1);
    y = pt(1,2);
    sigma_xx(igp) = -P*(L-x)*y/I;
end
varname{1}='X';
varname{2}='Y';
varname{3}='sigma_xx';
value=[Q(:,1) Q(:,2) sigma_xx'];
% tecplotout(tri,'T3',Q,varname,value,'beam_exact_stress.dat');

% Remove used memory
clear coord; clear Q; clear W; clear J; clear stress_gp;

% --------------------------------------------------------------------
%                           END OF THE PROGRAM
% --------------------------------------------------------------------
